local E = AleaUI
local CT = E:Module("Cooldown")
local UF = E:Module("UnitFrames")

local ICON_SIZE = 36 --the normal size for an icon (don't change this)
local FONT_SIZE = 18 --the base font size to use at a scale of 1
local MIN_SCALE = 0.2 --the minimum scale we want to show cooldown counts at, anything below this will be hidden
local MIN_DURATION = 2.5 --the minimum duration to show cooldown text for

local floor = math.floor
local min = math.min
local GetTime = GetTime

local threshold = 3

local TimeColors = {
	[0] = '|cfffefefe',
	[1] = '|cfffefefe',
	[2] = '|cfffefefe',
	[3] = '|cfffefefe',
	[4] = '|cfffe0000',
	[5] = '|cffff4c4c',
	[6] = '|cffff9999',
}

E.TimeFormats = {
	[0] = { '%.0fд', '%.0fд', '%.0fд' },
	[1] = { '%.0fч', '%.0fч', '%.0fч' },
	[2] = { '%.0fм', '%.0fм', '%.0fм' },
	[3] = { '%.0fс', '%.0f', '%.0f' },
	[4] = { '%.1fс', '%.1f', '%d' },
	[5] = { '%.1fс', '%.1f', '%d' },
	[6] = { '%.1fс', '%.1f', '%d' },
}

local defaults = {
	
	time_format = 1,
	
	show_millisec = true,
	
	point = 'TOPLEFT',
	offsetX = -1,
	offsetY = -1,
	
	font = E.media.default_font_name,
	fontSize = 18,
	fontOutline = 'OUTLINE',	
}

E.default_settings.cooldown = defaults

local JustifyByPoint = {
	['TOPLEFT'] = { 'LEFT', 'TOP' },
	['TOPRIGHT'] = { 'RIGHT', 'TOP' },
	['BOTTOMRIGHT'] = { 'RIGHT', 'BOTTOM' },
	['BOTTOMLEFT'] = { 'LEFT', 'BOTTOM' },
	['CENTER'] = { 'CENTER', 'CENTER' },
	['BOTTOM'] = { 'CENTER', 'BOTTOM' },
	['TOP'] = { 'CENTER', 'TOP' },
}

		
local fullCooldownList = {}

local DAY, HOUR, MINUTE = 86400, 3600, 60 --used for calculating aura time text
local DAYISH, HOURISH, MINUTEISH = HOUR * 23.5, MINUTE * 59.5, 59.5 --used for caclculating aura time at transition points
local HALFDAYISH, HALFHOURISH, HALFMINUTEISH = DAY/2 + 0.5, HOUR/2 + 0.5, MINUTE/2 + 0.5 --used for calculating next update times

-- will return the the value to display, the formatter id to use and calculates the next update for the Aura
function E:GetTimeInfo(s, threshhold)
	if s < MINUTE then
		if s >= threshhold then
			return ceil(s), 3, 0.51
		elseif s >= 2 then
			return E.db.cooldown.show_millisec and s or ceil(s), 6, 0.51
		elseif s >= 1 then
			return E.db.cooldown.show_millisec and s or ceil(s), 5, 0.51
		else
			return E.db.cooldown.show_millisec and s or ceil(s), 4, 0.51
		end
	elseif s < HOUR then
		local minutes = floor((s/MINUTE)+.5)
		return ceil(s / MINUTE), 2, minutes > 1 and (s - (minutes*MINUTE - HALFMINUTEISH)) or (s - MINUTEISH)
	elseif s < DAY then
		local hours = floor((s/HOUR)+.5)
		return ceil(s / HOUR), 1, hours > 1 and (s - (hours*HOUR - HALFHOURISH)) or (s - HOURISH)
	else
		local days = floor((s/DAY)+.5)
		return ceil(s / DAY), 0,  days > 1 and (s - (days*DAY - HALFDAYISH)) or (s - DAYISH)
	end
end

local function Cooldown_OnUpdate(cd, elapsed)
	if cd.nextUpdate > 0 then
		cd.nextUpdate = cd.nextUpdate - elapsed
		return
	end

	local remain = cd.duration - (GetTime() - cd.start)

	if remain > 0.05 then
		if (cd.fontScale * cd:GetEffectiveScale() / E.UIParent:GetScale()) < MIN_SCALE then
			cd.text:SetText('')
			cd.nextUpdate = 500
		else
			local timervalue, formatid
			timervalue, formatid, cd.nextUpdate = E:GetTimeInfo(remain, threshold)		
			cd.text:SetFormattedText(("%s%s|r"):format(TimeColors[formatid], E.TimeFormats[formatid][E.db.cooldown.show_millisec and 2 or 3]), timervalue)
		end
	else
		CT:Cooldown_StopTimer(cd)
	end
end

function CT:Cooldown_OnSizeChanged(cd, width, height)
	local fontScale = floor(width +.5) / ICON_SIZE
	local override = cd:GetParent():GetParent().SizeOverride
	if override then 
		fontScale = override / E.db.cooldown.fontSize  
	end
	
	if fontScale > 1 then
		fontScale = 1
	end
	
	if fontScale == cd.fontScale then
		return
	end
	
	cd.fontScale = fontScale
	if fontScale < MIN_SCALE and not override then
		cd:Hide()
	else

		cd:Show()
		cd.text:SetFont(E:GetFont(E.db.cooldown.font), fontScale * E.db.cooldown.fontSize, E.db.cooldown.fontOutline)
	
		if cd.enabled then
			self:Cooldown_ForceUpdate(cd)
		end
	end
end

function CT:Cooldown_ForceUpdate(cd)
	cd.nextUpdate = 0
	cd:Show()
end

function CT:Cooldown_StopTimer(cd)
	cd.enabled = nil
	cd:Hide()
end

function CT:CreateCooldownTimer(parent)
	local scaler = CreateFrame('Frame', nil, parent)
	scaler:SetAllPoints()

	local timer = CreateFrame('Frame', nil, scaler); timer:Hide()
	timer:SetAllPoints()
	timer:SetScript('OnUpdate', Cooldown_OnUpdate)

	local text = timer:CreateFontString(nil, 'OVERLAY')
	text:SetSize(50, 50)
	text:SetPoint(E.db.cooldown.point, timer, E.db.cooldown.point, E.db.cooldown.offsetX, E.db.cooldown.offsetY)
	text:SetJustifyH(JustifyByPoint[E.db.cooldown.point][1])
	text:SetJustifyV(JustifyByPoint[E.db.cooldown.point][2])
	
	timer.text = text

	self:Cooldown_OnSizeChanged(timer, parent:GetSize())
	parent:SetScript('OnSizeChanged', function(_, ...) self:Cooldown_OnSizeChanged(timer, ...) end)

	parent.timer = timer
	
	fullCooldownList[parent] = true
	
	return timer
end

function CT:OnSetCooldown(start, duration, charges, maxCharges)
	local button = self:GetParent()
	if self.noCooldownCount then return end
	
	if start > 0 and duration > MIN_DURATION then
		local timer = self.timer or CT:CreateCooldownTimer(self)
		timer.start = start
		timer.duration = duration
		timer.enabled = true
		timer.nextUpdate = 0
		if timer.fontScale >= MIN_SCALE then timer:Show() end
	else
		local timer = self.timer
		if timer then
			CT:Cooldown_StopTimer(timer)
		end
	end
	

	if self.timer then
		if charges and charges > 0 then
			self.timer:SetAlpha(0)
		else
			self.timer:SetAlpha(1)	
		end
	end
end

local function cooldown_Update(self)
	local button = self:GetParent()
	local action = button.action
	
	local start, duration, enable = GetActionCooldown(action)
	local charges, maxCharges, chargeStart, chargeDuration = GetActionCharges(action)
	
	CT.OnSetCooldown(self, start, duration, charges, maxCharges)
end

--- ACTION BUTTONS



local ActionBarButtonEventsFrame = _G['ActionBarButtonEventsFrame']
if not ActionBarButtonEventsFrame then return end

--[[ cooldown timer updating ]]--

local active = {}

local function cooldown_OnShow(self)
	active[self] = true
end

local function cooldown_OnHide(self)
	active[self] = nil
end

--returns true if the cooldown timer should be updated and false otherwise
local function cooldown_ShouldUpdateTimer(self, start, duration, charges, maxCharges)
	local timer = self.timer

	return not(
		timer 
		and timer.start == start 
		and timer.duration == duration 
		and timer.charges == charges 
		and timer.maxCharges == maxCharges
	)
end

local function cooldown_Update(self)
	local button = self:GetParent()
	local action = button.action
	
	local start, duration, enable = GetActionCooldown(action)
	local charges, maxCharges, chargeStart, chargeDuration = GetActionCharges(action)
	
	CT.OnSetCooldown(self, start, duration, charges, maxCharges)
end

do
	local watcher = CreateFrame('Frame')

	watcher:Hide()

	watcher:SetScript('OnEvent', function(self, event)
		for cooldown in pairs(active) do
			cooldown_Update(cooldown)
		end
	end)

	watcher:RegisterEvent('ACTIONBAR_UPDATE_COOLDOWN')
end

hooksecurefunc(getmetatable(_G['ActionButton1Cooldown']).__index, 'SetCooldown', CT.OnSetCooldown)	

local function actionButton_Register(frame)
	local cooldown = frame.cooldown
	if cooldown.noCooldownCount then return end
	
	if not cooldown.hooked then
		cooldown:HookScript('OnShow', cooldown_OnShow)
		cooldown:HookScript('OnHide', cooldown_OnHide)
		cooldown.hooked = true
	end
end

if ActionBarButtonEventsFrame.frames then
	for i, frame in pairs(ActionBarButtonEventsFrame.frames) do
		actionButton_Register(frame)
	end
end
hooksecurefunc('ActionBarButtonEventsFrame_RegisterFrame', actionButton_Register)


function CT:UpdateSettings()
	
	for cooldown in pairs(fullCooldownList) do
		if cooldown.timer then
			
			cooldown.timer.fontScale = nil
			CT:Cooldown_OnSizeChanged(cooldown.timer, cooldown:GetSize())
			
			cooldown.timer.text:ClearAllPoints()
			cooldown.timer.text:SetPoint(E.db.cooldown.point, cooldown.timer, E.db.cooldown.point, E.db.cooldown.offsetX, E.db.cooldown.offsetY)
			cooldown.timer.text:SetJustifyH(JustifyByPoint[E.db.cooldown.point][1])
			cooldown.timer.text:SetJustifyV(JustifyByPoint[E.db.cooldown.point][2])
	
		end
	end
	
end

E.GUI.args.Cooldown = {
	name = "Текст перезарядки",
	type = "group",
	order = 5,
	args = {},
}

E.GUI.args.Cooldown.args.ShowMillisec = {
	name = 'Показывать миллисекунды',
	type = 'toggle',
	order = 1, width = 'full',
	set = function()
		E.db.cooldown.show_millisec = not E.db.cooldown.show_millisec
		CT:UpdateSettings()
	end,
	get = function()
		return E.db.cooldown.show_millisec
	end,	
}

E.GUI.args.Cooldown.args.fontTemplate = {
	name = "Стиль",
	type = "group",
	order = 5,
	embend = true,
	args = {},
}
E.GUI.args.Cooldown.args.fontTemplate.args.font = {	
	name = "Шрифт",
	order = 6,
	type = "font",
	values = E.GetFontList,
	set = function(self, value)
		E.db.cooldown.font = value
		CT:UpdateSettings()
	end,
	get = function(self)
		return E.db.cooldown.font
	end,
}
E.GUI.args.Cooldown.args.fontTemplate.args.fontOutline = {	
	name = "Контур",
	order = 7,
	type = "dropdown",
	values = {			
		["NONE"] = NO,
		["OUTLINE"] = "OUTLINE",
	},
	set = function(self, value)
		E.db.cooldown.fontOutline = value
		CT:UpdateSettings()
	end,
	get = function(self)
		return E.db.cooldown.fontOutline
	end,
}
E.GUI.args.Cooldown.args.fontTemplate.args.fontSize = {	
	name = "Размер",
	order = 8,
	type = "slider",
	min = 1, max = 32, step = 1,
	set = function(self, value)
		E.db.cooldown.fontSize = value
		CT:UpdateSettings()
	end,
	get = function(self)
		return E.db.cooldown.fontSize
	end,
}

E.GUI.args.Cooldown.args.fontPoint = {
	name = "Расположение",
	type = "group",
	order = 4,
	embend = true,
	args = {},
}

E.GUI.args.Cooldown.args.fontPoint.args.xOffset = {
	name = 'Отступ по горизонтали',
	order = 3,
	type = 'slider',
	min = -50, max = 50, step = 1,
	set = function(info, value)
		E.db.cooldown.offsetX = value
		CT:UpdateSettings()
	end,
	get = function(info)
		return E.db.cooldown.offsetX
	end,
}
E.GUI.args.Cooldown.args.fontPoint.args.yOffset = {
	name = 'Отступ по вертикали',
	order = 4,
	type = 'slider',
	min = -50, max = 50, step = 1,
	set = function(info, value)
		E.db.cooldown.offsetY = value
		CT:UpdateSettings()
	end,
	get = function(info)
		return E.db.cooldown.offsetY
	end,
}		
E.GUI.args.Cooldown.args.fontPoint.args.point = {
	name = 'Точка фиксации',
	order = 5,
	type = 'dropdown',
	values = {
		['TOPLEFT'] = 'TOPLEFT',
		['TOPRIGHT'] = 'TOPRIGHT',
		['BOTTOMRIGHT'] = 'BOTTOMRIGHT',
		['BOTTOMLEFT'] = 'BOTTOMLEFT',
		['CENTER'] = 'CENTER',
		['BOTTOM'] = 'BOTTOM',
		['TOP'] = 'TOP',
	},
	set = function(info, value)
		E.db.cooldown.point = value
		CT:UpdateSettings()
	end,
	get = function(info)
		return E.db.cooldown.point
	end,
}
